//+------------------------------------------------------------------+
//|                                                OptimizerBase.mqh |
//|                                      Copyright 2018, Dale Woods. |
//|                                      https://www.theforexguy.com |
//+------------------------------------------------------------------+
#property copyright "Copyright 2018, Dale Woods."
#property link      "https://www.theforexguy.com"
#property version   "1.00"


struct OptimizedInput {
   
   private:
            
      bool 
         _paramsSet,
         _optimzing;
      
      int 
         _precision,
         _stepPrecision;
      
         
      double
         _currentStartPos;
         
      void DumpInputs(const double &usingInputs[]) const;  
   
      
   public: 
      double 
         startValue,
         stepValue,
         endValue;
         
      
      OptimizedInput() { _paramsSet = false; Reset(); }   
      
         
      OptimizedInput(const double inStart, const double inStep, const double inEnd, const bool enableOptimization = true) { 
         Reset();
         
         if( inEnd < inStart || inStep > inEnd ) { 
            printf("[%s ERROR]: Invalid inputs passed", __FUNCTION__);
            _paramsSet = false; 
            return;
         }
         
         startValue = inStart;
         stepValue = inStep;
         endValue = inEnd;     
         
         _paramsSet = true; 
         _optimzing = enableOptimization;
         
         _precision = 2;
         _stepPrecision = 1;
         
         _currentStartPos = StartValue();
      }
      
      
      
      void Reset() {
         startValue = 0;
         stepValue = 0;
         endValue = 0;
         
         _paramsSet = false; 
         _optimzing = true;
         
         _precision = 2;
         _stepPrecision = 1;
         
         _currentStartPos = StartValue();
      }   
      
      
      
      void operator = (const OptimizedInput &source) {
         
         startValue  = source.startValue;
         stepValue   = source.stepValue;
         endValue    = source.endValue;
         
         _paramsSet  = source.AreParamsSet();
         _optimzing = source.Optimize();         
         _precision = source.Precision();
         _stepPrecision = source.StepPrecision();
         
         _currentStartPos  = source.CurrentStartPos();
      } 
      
      
      
      bool operator == (const OptimizedInput &source) {
         bool isEqual = true;
         
         isEqual &= (StartValue() == source.StartValue());
         isEqual &= (StepValue()  == source.StepValue());
         isEqual &= (EndValue()   == source.EndValue());         
         
         return isEqual;
      }
      
            
      
      bool AreParamsSet() const                    { return _paramsSet; } 
      
      bool Optimize() const                        { return _optimzing; }
      void Optimize(const bool shouldOptimize)     { _optimzing = shouldOptimize; }
      
      int Precision() const                        { return _precision; }
      void Precision(const int setPrecision)       { _precision = setPrecision; }
      
      int StepPrecision() const                    { return _stepPrecision; }
      void StepPrecision(const int setPrecision)   { _stepPrecision = setPrecision; }     
      
      double StartValue() const                    { return NormalizeDouble(startValue, _precision); }
      double StepValue() const                     { return NormalizeDouble(stepValue, _stepPrecision); }
      double EndValue() const                      { return NormalizeDouble(endValue, _precision); }
      
      double CurrentStartPos() const               { return NormalizeDouble(_currentStartPos, _precision); } 
      void IncrementStartPos()                     { _currentStartPos = CurrentStartPos() + StepValue(); } 
};



//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class OptimizerBase {

   private:      
      int _inputCount;

      OptimizedInput _inputFields[];
      void ResetInputs(double &arr[]);
      
      
      void DumpInputs(const double &usingInputs[]) const;
      
   public:
      OptimizerBase();
      ~OptimizerBase();      
      
      void AddInput(OptimizedInput &inInput);
      void RunOptimizer();
      
      void RunStrategyUsingInputs(const double &usingInputs[]);
};



OptimizerBase::OptimizerBase() {
   
   // Init variables
   _inputCount = 0;   
}



OptimizerBase::~OptimizerBase() {

}



/*
   Must be passed in correct order
*/
void OptimizerBase::AddInput(OptimizedInput &inInput) {
   
   if( !inInput.AreParamsSet() ) { 
      printf("[%s ERROR]: Passed Optimzied Input struct is empty", __FUNCTION__);
      return;
   }
  
   // Append to array 
   const int arrSize = ArraySize(_inputFields);
   ArrayResize(_inputFields, arrSize+1);
   _inputFields[arrSize] = inInput;
   
   _inputCount++;
}



/*
void OptimizerBase::RunOptimizer(const int atCandle) {

   // Array to hold inputs for each pass
   double iteratedInputs[];   
   
   ArrayResize(iteratedInputs, _inputCount);
   
   int count = 0;
   
   //for(int fieldNum = 0; fieldNum < _inputCount; fieldNum++) { 
      int fieldNum = 0;
      // Loop that cycles through iterations of focused input
      for( double fieldItr = _inputFields[fieldNum].StartValue(); fieldItr < _inputFields[fieldNum].EndValue() + _inputFields[fieldNum].StepValue(); fieldItr += _inputFields[fieldNum].StepValue() ) {
         
         for( int subInput = ArraySize(_inputFields)-1; subInput > fieldNum; subInput-- ) {
            
            ResetInputs(iteratedInputs);
            
            iteratedInputs[fieldNum] = fieldItr;
            
            for( double subfieldItr = _inputFields[subInput].StartValue(); subfieldItr < _inputFields[subInput].EndValue() + _inputFields[subInput].StepValue(); subfieldItr += _inputFields[subInput].StepValue() ) {
               
               
               iteratedInputs[subInput] = subfieldItr;
               
               DumpInputs(iteratedInputs);
            }      
         }
      }   
   //}
}*/





void OptimizerBase::RunOptimizer() {
   
   // Array to hold inputs for each pass
   double iteratedInputs[];   
   
   ArrayResize(iteratedInputs, _inputCount);
   
   int count = 0;
   
   for(int fieldNum = 0; fieldNum < _inputCount; fieldNum++) {      
      
      // Skip this input if optimize is not enabled   
      if( !_inputFields[fieldNum].Optimize() ) { continue; }
      
      int level2pass = 1;
      
      // Loop that cycles through iterations of focused input
      for( double fieldItr = _inputFields[fieldNum].CurrentStartPos(); fieldItr < _inputFields[fieldNum].EndValue() + _inputFields[fieldNum].StepValue(); fieldItr += _inputFields[fieldNum].StepValue() ) {
               
         int level3pass = 2;
         //if( level2pass < fieldNum ) { level2pass++; continue; }  
         
         // Loop that cycles through iterations of unfocused inputs against current focused input interation
         //for(int subFieldNum = 0; subFieldNum < _inputCount; subFieldNum++) {  
         
         
         
         for(int subFieldNum = _inputCount-1; subFieldNum >= 0; subFieldNum--) {  
            if( subFieldNum == fieldNum ) { continue; }
                    
            //if( j == fieldNum ) { level3pass++; continue; }          
            //if( level3pass < level2pass ) { level3pass++; continue; }  
            
            double level4pass = 3;
                      
            for( double subfieldItr = _inputFields[subFieldNum].StartValue(); subfieldItr < _inputFields[subFieldNum].EndValue() + _inputFields[subFieldNum].StepValue(); subfieldItr += _inputFields[subFieldNum].StepValue() ) {   
               
               //printf("level 3 pass %i | subPass pass %i | input pass %i", level3pass, subPass, fieldNum);
               
               if( IsStopped() ) { return; }
               
               //if( level4pass < level3pass ) { level4pass++; continue; }
               
               ResetInputs(iteratedInputs);
            
               // Update optimized input
               iteratedInputs[fieldNum] = fieldItr;               
               iteratedInputs[subFieldNum] = subfieldItr; 
               
               level4pass++;              
               count++;
               
               //_inputFields[subFieldNum].IncrementStartPos();
               
               DumpInputs(iteratedInputs);    
               
                        
            }// for subF
            
            //_inputFields[subFieldNum].startValue = subStartValue;
            
            level3pass++;             
         }// for j
         
                 
      }// for fieldItr
      level2pass++;  
   
   }//end for fieldNum 
   
   printf("iterations generated = %i", count);
}










/*
void OptimizerBase::RunOptimizer(const int atCandle) {
   
   // Array to hold inputs for each pass
   double iteratedInputs[];   
   
   ArrayResize(iteratedInputs, _inputCount);
   
   int count = 0;
   
   for(int fieldNum = 0; fieldNum < _inputCount; fieldNum++) {      
      
      // Skip this input if optimize is not enabled   
      if( !_inputFields[fieldNum].Optimize() ) { continue; }
      
      int level2pass = 1;
      
      // Loop that cycles through iterations of focused input
      for( double fieldItr = _inputFields[fieldNum].StartValue(); fieldItr < _inputFields[fieldNum].EndValue() + _inputFields[fieldNum].StepValue(); fieldItr += _inputFields[fieldNum].StepValue() ) {
               
         int level3pass = 2;
         if( level2pass < fieldNum ) { level2pass++; continue; }  
         
         // Loop that cycles through iterations of unfocused inputs against current focused input interation
         for( int j = fieldNum; j < _inputCount; j++ )   {
                    
            //if( j == fieldNum ) { level3pass++; continue; }          
            if( level3pass < level2pass ) { level3pass++; continue; }  
            
            double level4pass = 3;
                       
            for( double subfieldItr = _inputFields[j].StartValue(); subfieldItr < _inputFields[j].EndValue() + _inputFields[j].StepValue(); subfieldItr += _inputFields[j].StepValue() ) {   
               
               //printf("level 3 pass %i | subPass pass %i | input pass %i", level3pass, subPass, fieldNum);
               
               if( IsStopped() ) { return; }
               
               if( level4pass < level3pass ) { level4pass++; continue; }
               
               ResetInputs(iteratedInputs);
            
               // Update optimized input
               iteratedInputs[fieldNum] = fieldItr;               
               iteratedInputs[j] = subfieldItr; 
               
               level4pass++;              
               count++;
               
               DumpInputs(iteratedInputs);             
            }// for subF
            
            level3pass++;             
         }// for j
         
                 
      }// for fieldItr
      level2pass++;  
      // Update start point of level l iteration to avoid repeat passes       
      //_inputFields[fieldNum].IncrementStartPos();   
   
   }//end for fieldNum 
   
   printf("iterations generated = %i", count);
}
*/



void OptimizerBase::ResetInputs(double &arr[]) { 
   for( int i = 0; i < _inputCount; i++ ) { arr[i] = _inputFields[i].CurrentStartPos(); }
}


void OptimizerBase::RunStrategyUsingInputs(const double &usingInputs[]) {

}



void OptimizerBase::DumpInputs(const double &usingInputs[]) const { 
   
   string concatInputs = "";
   
   for(int i = 0; i < ArraySize(usingInputs); i++ ) {
      //string s = StringFormat("input %i = %.3f | ", i, usingInputs[i]);
      string s = StringFormat("%.3f ", usingInputs[i]);
      StringConcatenate(concatInputs, concatInputs, s);    
   }
   
   printf("%s", concatInputs);

}